# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Andreas Kosubek <andreas.kosubek@ama.gv.at>
#          Bernhard Mallinger <bernhard.mallinger@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2023 Agrar Markt Austria
# Copyright (C) 2024 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import json
from typing import Any, Dict, Optional, Tuple

from ... import ast, values
from ..evaluator import Evaluator, handle

COMPARISON_OP_MAP = {
    ast.ComparisonOp.EQ: "=",
    ast.ComparisonOp.NE: "<>",
    ast.ComparisonOp.LT: "<",
    ast.ComparisonOp.LE: "<=",
    ast.ComparisonOp.GT: ">",
    ast.ComparisonOp.GE: ">=",
}


ARITHMETIC_OP_MAP = {
    ast.ArithmeticOp.ADD: "+",
    ast.ArithmeticOp.SUB: "-",
    ast.ArithmeticOp.MUL: "*",
    ast.ArithmeticOp.DIV: "/",
}

SPATIAL_COMPARISON_OP_MAP = {
    ast.SpatialComparisonOp.INTERSECTS: "ANYINTERACT",
    ast.SpatialComparisonOp.DISJOINT: "DISJOINT",
    ast.SpatialComparisonOp.CONTAINS: "CONTAINS",
    ast.SpatialComparisonOp.WITHIN: "INSIDE",
    ast.SpatialComparisonOp.TOUCHES: "TOUCH",
    ast.SpatialComparisonOp.CROSSES: "OVERLAPBDYDISJOINT",
    ast.SpatialComparisonOp.OVERLAPS: "OVERLAPBDYINTERSECT",
    ast.SpatialComparisonOp.EQUALS: "EQUAL",
}


class OracleSQLEvaluator(Evaluator):
    bind_variables: Dict[str, Any]

    def __init__(self, attribute_map: Dict[str, str], function_map: Dict[str, str]):
        self.attribute_map = attribute_map
        self.function_map = function_map

        self.with_bind_variables = False
        self.bind_variables = {}
        # Counter for bind variables
        self.b_cnt = 0

    @handle(ast.Not)
    def not_(self, node, sub):
        return f"NOT {sub}"

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        return f"({lhs} {node.op.value} {rhs})"

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        if self.with_bind_variables:
            self.bind_variables[f"{lhs}_{self.b_cnt}"] = rhs
            sql = f"({lhs} {COMPARISON_OP_MAP[node.op]} :{lhs}_{self.b_cnt})"
            self.b_cnt += 1
        else:
            sql = f"({lhs} {COMPARISON_OP_MAP[node.op]} {rhs})"
        return sql

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        if self.with_bind_variables:
            self.bind_variables[f"{lhs}_high_{self.b_cnt}"] = high
            self.bind_variables[f"{lhs}_low_{self.b_cnt}"] = low
            sql = (
                f"({lhs} {'NOT ' if node.not_ else ''}BETWEEN "
                f":{lhs}_low_{self.b_cnt} AND :{lhs}_high_{self.b_cnt})"
            )
            self.b_cnt += 1
        else:
            sql = f"({lhs} {'NOT ' if node.not_ else ''}BETWEEN " f"{low} AND {high})"
        return sql

    @handle(ast.Like)
    def like(self, node, lhs):
        pattern = node.pattern
        if node.wildcard != "%":
            pattern = pattern.replace(node.wildcard, "%")
        if node.singlechar != "_":
            pattern = pattern.replace(node.singlechar, "_")

        if self.with_bind_variables:
            self.bind_variables[f"{lhs}_{self.b_cnt}"] = pattern
            sql = f"{lhs} {'NOT ' if node.not_ else ''}LIKE "
            sql += f":{lhs}_{self.b_cnt} ESCAPE '{node.escapechar}'"

        else:
            sql = f"{lhs} {'NOT ' if node.not_ else ''}LIKE "
            sql += f"'{pattern}' ESCAPE '{node.escapechar}'"

        return sql

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return f"{lhs} {'NOT ' if node.not_ else ''}IN ({', '.join(options)})"

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return f"{lhs} IS {'NOT ' if node.not_ else ''}NULL"

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        param = f"mask={SPATIAL_COMPARISON_OP_MAP[node.op]}"
        func = f"SDO_RELATE({lhs}, {rhs}, '{param}') = 'TRUE'"
        return func

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        geo_json = json.dumps(
            {
                "type": "Polygon",
                "coordinates": [
                    [
                        [node.minx, node.miny],
                        [node.minx, node.maxy],
                        [node.maxx, node.maxy],
                        [node.maxx, node.miny],
                        [node.minx, node.miny],
                    ]
                ],
            }
        )
        srid = 4326
        param = "mask=ANYINTERACT"

        if self.with_bind_variables:
            self.bind_variables[f"geo_json_{self.b_cnt}"] = geo_json
            self.bind_variables[f"srid_{self.b_cnt}"] = srid
            geom_sql = (
                f"SDO_UTIL.FROM_JSON(geometry => :geo_json_{self.b_cnt}, "
                f"srid => :srid_{self.b_cnt})"
            )
            self.b_cnt += 1
        else:
            geom_sql = (
                f"SDO_UTIL.FROM_JSON(geometry => '{geo_json}', " f"srid => {srid})"
            )

        sql = f"SDO_RELATE({lhs}, {geom_sql}, '{param}') = 'TRUE'"
        return sql

    @handle(ast.Attribute)
    def attribute(self, node: ast.Attribute):
        return f"{self.attribute_map[node.name]}"

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node: ast.Arithmetic, lhs, rhs):
        op = ARITHMETIC_OP_MAP[node.op]
        return f"({lhs} {op} {rhs})"

    @handle(ast.Function)
    def function(self, node, *arguments):
        func = self.function_map[node.name]
        return f"{func}({','.join(arguments)})"

    @handle(*values.LITERALS)
    def literal(self, node):
        if isinstance(node, str):
            return f"'{node}'"
        else:
            return node

    @handle(values.Geometry)
    def geometry(self, node: values.Geometry):
        # TODO Read CRS information from
        #      node and translate to SRID
        srid = 4326
        geo_json = json.dumps(node.geometry)
        print(geo_json)
        if self.with_bind_variables:
            self.bind_variables[f"geo_json_{self.b_cnt}"] = geo_json
            self.bind_variables[f"srid_{self.b_cnt}"] = srid
            sql = (
                f"SDO_UTIL.FROM_JSON(geometry => :geo_json_{self.b_cnt}, "
                f"srid => :srid_{self.b_cnt})"
            )
            self.b_cnt += 1
        else:
            sql = f"SDO_UTIL.FROM_JSON(geometry => '{geo_json}', " f"srid => {srid})"
        return sql

    @handle(values.Envelope)
    def envelope(self, node: values.Envelope):
        # TODO Read CRS information from
        #      node and translate to SRID
        srid = 4326
        geo_json = json.dumps(node.geometry)
        if self.with_bind_variables:
            self.bind_variables[f"geo_json_{self.b_cnt}"] = geo_json
            self.bind_variables[f"srid_{self.b_cnt}"] = srid
            sql = (
                f"SDO_UTIL.FROM_JSON(geometry => :geo_json_{self.b_cnt}, "
                f"srid => :srid_{self.b_cnt})"
            )
            self.b_cnt += 1
        else:
            sql = f"SDO_UTIL.FROM_JSON(geometry => '{geo_json}', " f"srid => {srid})"
        return sql


def to_sql_where(
    root: ast.Node,
    field_mapping: Dict[str, str],
    function_map: Optional[Dict[str, str]] = None,
) -> str:
    orcle = OracleSQLEvaluator(field_mapping, function_map or {})
    orcle.with_bind_variables = False
    return orcle.evaluate(root)


def to_sql_where_with_bind_variables(
    root: ast.Node,
    field_mapping: Dict[str, str],
    function_map: Optional[Dict[str, str]] = None,
) -> Tuple[str, Dict[str, Any]]:
    orcle = OracleSQLEvaluator(field_mapping, function_map or {})
    orcle.with_bind_variables = True
    orcle.bind_variables = {}
    return orcle.evaluate(root), orcle.bind_variables



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

"""Install pygeofilter."""

import os
import os.path

from setuptools import find_packages, setup

# don't install dependencies when building win readthedocs
on_rtd = os.environ.get("READTHEDOCS") == "True"

# use README.md for project long_description
with open("README.md") as f:
    readme = f.read()

description = (
    "pygeofilter is a pure Python parser implementation of OGC filtering standards"
)

setup(
    name="pygeofilter",
    description=description,
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Fabian Schindler",
    author_email="fabian.schindler@eox.at",
    url="https://github.com/geopython/pygeofilter",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=(
        [
            "dateparser",
            "lark",
            "pygeoif>=1.0.0",
            "dataclasses;python_version<'3.7'",
        ]
        if not on_rtd
        else []
    ),
    extras_require={
        "backend-django": ["django"],
        "backend-sqlalchemy": ["geoalchemy2", "sqlalchemy"],
        "backend-native": ["shapely"],
        "backend-elasticsearch": ["elasticsearch", "elasticsearch-dsl"],
        "backend-opensearch": ["opensearch-py", "opensearch-dsl"],
        "fes": ["pygml>=0.2"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: GIS",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    tests_require=["pytest"],
)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from pygeofilter import ast
from pygeofilter.backends.optimize import optimize
from pygeofilter.parsers.ecql import parse


def test_not():
    result = optimize(parse("NOT 1 > 2"))
    assert result == ast.Include(False)

    result = optimize(parse("NOT 1 < 2"))
    assert result == ast.Include(True)


def test_combination():
    # reduce right hand side
    result = optimize(parse("attr = 1 AND 1 < 2"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # reduce left hand side
    result = optimize(parse("1 < 2 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # reduce left hand side
    result = optimize(parse("1 < 2 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # can' reduce
    result = optimize(parse("attr = 1 AND other = 2"))
    assert result == ast.And(
        ast.Equal(ast.Attribute("attr"), 1), ast.Equal(ast.Attribute("other"), 2)
    )

    # reduce AND to an INCLUDE if both sides evaluate to true
    result = optimize(parse("1 = 1 AND 2 = 2"))
    assert result == ast.Include(False)

    # reduce AND to an EXCLUDE if either side evaluates to false
    result = optimize(parse("attr = 1 AND 2 = 3"))
    assert result == ast.Include(True)
    result = optimize(parse("2 = 3 AND attr = 1"))
    assert result == ast.Include(True)
    result = optimize(parse("0 = 1 AND 2 = 3"))
    assert result == ast.Include(True)

    # reduce OR to INCLUDE if either side evaluates to true
    result = optimize(parse("attr = 1 OR 2 = 2"))
    assert result == ast.Include(False)
    result = optimize(parse("2 = 2 OR attr = 1"))
    assert result == ast.Include(False)

    # reduce OR to an EXCLUDE if both sides evaluate to false
    result = optimize(parse("1 = 2 AND 2 = 1"))
    assert result == ast.Include(True)


def test_comparison():
    # reduce less than
    result = optimize(parse("1 < 2 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # reduce greater than
    result = optimize(parse("2 > 1 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # reduce less or equal
    result = optimize(parse("1 <= 2 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # reduce greater or equal
    result = optimize(parse("2 >= 1 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # reduce not equal
    result = optimize(parse("2 <> 1 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)


def test_between():
    # allow reduction
    result = optimize(parse("5 BETWEEN 1 AND 6 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)
    result = optimize(parse("10 NOT BETWEEN 1 AND 6 AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # don't reduce if either lhs, low or high are uncertain
    result = optimize(parse("attr BETWEEN 1 AND 6"))
    assert result == ast.Between(ast.Attribute("attr"), 1, 6, False)
    result = optimize(parse("5 BETWEEN attr AND 6"))
    assert result == ast.Between(5, ast.Attribute("attr"), 6, False)
    result = optimize(parse("5 BETWEEN 1 AND attr"))
    assert result == ast.Between(5, 1, ast.Attribute("attr"), False)


def test_like():
    # allow reduction
    result = optimize(parse("'This is a test' LIKE 'This is %' AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)
    result = optimize(parse("'This is a test' LIKE 'This is . test' AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)

    # don't reduction when an attribute is referenced
    result = optimize(parse("attr LIKE 'This is %'"))
    assert result == ast.Like(
        ast.Attribute("attr"), "This is %", False, "%", ".", "\\", False
    )


def test_in():
    # allow reduction when the left hand side and all options
    # are certain
    result = optimize(parse("1 IN (1, 2, 3) AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)
    result = optimize(parse("5 NOT IN (1, 2, 3) AND attr = 1"))
    assert result == ast.Equal(ast.Attribute("attr"), 1)
    # don't allow reduction if either left hand side or either option
    # is uncertain
    result = optimize(parse("attr IN (1, 2, 3)"))
    assert result == ast.In(ast.Attribute("attr"), [1, 2, 3], False)
    result = optimize(parse("1 IN (attr, 2, 3)"))
    assert result == ast.In(1, [ast.Attribute("attr"), 2, 3], False)


def test_temporal():
    # TODO
    pass


def test_array():
    # TODO
    pass


def test_spatial():
    # TODO
    pass


def test_arithmetic():
    # test possible optimizations
    result = optimize(parse("attr = 10 + 10"))
    assert result == ast.Equal(ast.Attribute("attr"), 20)

    result = optimize(parse("attr = 30 - 10"))
    assert result == ast.Equal(ast.Attribute("attr"), 20)

    result = optimize(parse("attr = 10 * 2"))
    assert result == ast.Equal(ast.Attribute("attr"), 20)

    result = optimize(parse("attr = 40 / 2"))
    assert result == ast.Equal(ast.Attribute("attr"), 20)

    # test imppossible optimizations
    result = optimize(parse("attr = other + 10"))
    assert result == ast.Equal(
        ast.Attribute("attr"),
        ast.Add(ast.Attribute("other"), 10),
    )

    result = optimize(parse("attr = other - 10"))
    assert result == ast.Equal(
        ast.Attribute("attr"),
        ast.Sub(ast.Attribute("other"), 10),
    )

    result = optimize(parse("attr = other * 2"))
    assert result == ast.Equal(
        ast.Attribute("attr"),
        ast.Mul(ast.Attribute("other"), 2),
    )

    result = optimize(parse("attr = other / 2"))
    assert result == ast.Equal(
        ast.Attribute("attr"),
        ast.Div(ast.Attribute("other"), 2),
    )


def test_function():
    def myadder(a, b):
        return a + b

    result = optimize(parse("attr = myadder(1, 2)"), {"myadder": myadder})
    assert result == ast.Equal(
        ast.Attribute("attr"),
        3,
    )

    # can't optimize a function referencing an attribute
    result = optimize(parse("attr = myadder(other, 2)"), {"myadder": myadder})
    assert result == ast.Equal(
        ast.Attribute("attr"), ast.Function("myadder", [ast.Attribute("other"), 2])
    )
    # can't optimize a function with a nested reference to an attribute
    result = optimize(parse("attr = myadder(other + 2, 2)"), {"myadder": myadder})
    assert result == ast.Equal(
        ast.Attribute("attr"),
        ast.Function("myadder", [ast.Add(ast.Attribute("other"), 2), 2]),
    )

    # can't optimize an unknown functions
    result = optimize(parse("attr = unkown(1, 2)"), {"myadder": myadder})
    assert result == ast.Equal(
        ast.Attribute("attr"),
        ast.Function(
            "unkown",
            [
                1,
                2,
            ],
        ),
    )
